/*
 *  GuiLoader, a GuiXml loader library
 *  Copyright (c) 2006 Maxim Udushlivy
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef GUILOADER_GUILOADERIMPL_H
#define GUILOADER_GUILOADERIMPL_H

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <libintl.h>

#define GUILOADER_LIBRARY

#ifdef GUILOADER_SANDBOX
	#define G_DISABLE_DEPRECATED
	#define GDK_DISABLE_DEPRECATED
	#define GDK_PIXBUF_DISABLE_DEPRECATED
	#define PANGO_DISABLE_DEPRECATED
	#define GTK_DISABLE_DEPRECATED
#endif

#include "guiloader.h"

#define GUILOADER_VERSION_GUIXML (9)
#define GUI_NEEDS_UPGRADE (8)

#ifdef G_LOG_DOMAIN
	#undef G_LOG_DOMAIN
#endif
#define G_LOG_DOMAIN "GuiLoader"

#define GUI_LOADER_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE((obj), GUI_TYPE_LOADER, GuiLoaderPrivate))
#define Check(expr) if(!(expr)) CheckFailed(#expr, __FILE__, __LINE__); else

#define GUILOADER_VIRTUAL_BASE(proxy, vb)  \
  if((proxy->bases & vb) == vb)            \
    return;                                \
  else                                     \
    proxy->bases |= vb;

typedef enum { 
	vbGObject=        1, 
	vbGtkWidget=      2, 
	vbGtkContainer=   4 
} VirtualBases;

typedef enum { 
	CROW_PACK_SHRINK, 
	CROW_PACK_EXPAND_PADDING,
	CROW_PACK_EXPAND_WIDGET 
} CrowPackOptions;

typedef enum { nrNone, nrEntity, nrScalar, nrVector, nrLink } NodeRole;

typedef struct _GuiLoaderPrivate GuiLoaderPrivate;
typedef struct _Node Node;
typedef struct _PropertyFuncData PropertyFuncData;
typedef struct _Property Property;
typedef struct _Proxy Proxy;
typedef struct _CrowPoint CrowPoint;
typedef struct _DelayedAssignment DelayedAssignment;
typedef struct _PtrArray PtrArray;
typedef struct _CrowEmitter CrowEmitter;
typedef struct _ShortStr ShortStr;
typedef struct _CustomConstructor CustomConstructor;

typedef GObject * (*InstanceCreateFunc)(GuiLoader *, Node *);
typedef void (*InstanceInitFunc)(GObject *, Node *);
typedef void (*PropertyFunc)(PropertyFuncData *);
typedef void (*ProxyCreateFunc)(Proxy *);
typedef void (*ForEachNodeFunc)(Node *);

typedef const gchar * pconstgchar;

struct _GuiLoaderPrivate {
	gboolean preview;
	gboolean steal;
	GSList * stack;
	Node * model;
	int version;
	GHashTable * objects;
	GHashTable * connectors;
	GSList * delayed;
	GSList * strings;
	GHashTable * constructors;
};

struct _GuiLoaderInit {
	int refcounter;
	GHashTable * proxies;
};

struct _Node {
	NodeRole role;
	ShortStr * name;
	ShortStr * type;
	ShortStr * meta;
	ShortStr * text;
	Node * owner;
	GQueue * domain;
	Proxy * proxy;
	GObject * object;
};

struct _PropertyFuncData {
	GuiLoader * loader;
	GObject * object;
	const gchar * name;
	Node * node;	
};

struct _Property {
	const gchar * name;
	PropertyFunc assign;
	gboolean lazy;
};

struct _Proxy {
	GType type;
	GQueue * properties;
	InstanceCreateFunc createInstance;
	GQueue * initInstanceList;
	VirtualBases bases;
	int priority;
};

struct _CrowPoint {
	int x;
	int y;
};

struct _DelayedAssignment {
	GObject * object;
	Property * property;
	Node * node;
};

struct _PtrArray {
	gsize size;
	gpointer ptrs[1];
};

struct _CrowEmitter {
	ShortStr * signame;
	ShortStr * mapname;
	gboolean after;
};

struct _GuiConnector {
	GObject * object;
	CrowEmitter * emitter;
};

struct _ShortStr {
	union {
		struct {
			guint32 len : 16;
			guint32 refcounter : 8;
			guint32 ch : 8;
		};
		gchar buf[4];
	};
};

struct _CustomConstructor {
	GuiCustomConstructorFunc func;
	gpointer data;
	GDestroyNotify destroy;
};

void gui_loader_class_init(GuiLoaderClass * meta);
void gui_loader_init(GuiLoader * loader);
void gui_loader_finalize(GObject * object);

void CheckFailed(const gchar * expr, const gchar * file, unsigned line);
void FatalError(const gchar * message);

ShortStr * ShortStrNew(const gchar * str);
ShortStr * ShortStrNewLen(const gchar * str, gsize length);
ShortStr * ShortStrAlloc(gsize length);
gchar * ShortStrGetStr(ShortStr * string);
void ShortStrRef(ShortStr * string);
void ShortStrUnref(ShortStr * string);
ShortStr * ShortStrReplace(ShortStr * string, int pos, int len, const gchar * repstr);
guint ShortStrHash(gconstpointer string);
gboolean ShortStrEqual(gconstpointer s1, gconstpointer s2);

PtrArray * PtrArrayNew(gsize size);
void PtrArrayFree(PtrArray * array);
void PtrArrayNull(PtrArray * array);

GList * CreateListFromPtrArray(PtrArray * array);
pconstgchar * SplitString(const gchar * str, gchar ch, int * maxcount);
gsize SplitLength(pconstgchar * ptrs, int i);
void FreeSplitPtrs(pconstgchar * ptrs, int count);
gboolean BoolFromString(const gchar * str);
const gchar * Translate(const gchar * meta, const gchar * str, gsize * size);
gboolean StrEqual(const gchar * s1, const gchar * s2);

void InitInstance(GuiLoader * loader);
void DoneInstance(GuiLoader * loader);

void CustomConstructorFree(CustomConstructor * constructor);

void LoadModel(GuiLoader * loader, const gchar * data, gssize size);
void ForEachNode(Node * node, ForEachNodeFunc func);
void Upgrade8(Node * node);
void Upgrade8Rename(Node * node);
void Upgrade8Fix(Node * node);
void BindProxy(Node * node);
Node * NewNode();
void FreeModel(GuiLoader * loader);
void FreeNode(Node * node);
Node * FindNode(Node * node, const gchar * name);
ShortStr * AddSharedStr(GuiLoader * loader, ShortStr * string);

void LoadGui(GuiLoader * loader);
gint RootOrder(gconstpointer pa, gconstpointer pb, gpointer usedata);
void FreeGui(GuiLoader * loader);
gboolean RemoveObject(gpointer key, gpointer value, gpointer userdata);
gboolean RemoveConnector(gpointer key, gpointer value, gpointer userdata);
void MakeObject(GuiLoader * loader, Node * node, GObject * old);
void LoadObject(GuiLoader * loader, Node * node);
void AddConnector(GuiLoader * loader, GObject * object, CrowEmitter * emitter);
void AddDelayedAssignment(GuiLoader * loader, GObject * object, Property * property, Node * node);
GObject * FindObject(GuiLoader * loader, const gchar * name);

GuiLoaderInit * NewInit(void);
void FreeInit(GuiLoaderInit * init);
gboolean RemoveProxy(gpointer key, gpointer value, gpointer userdata);
void AddProxy(GuiLoaderInit * init, GType type, ProxyCreateFunc create, int priority);
Property * AddProperty(Proxy * proxy, const gchar * name, PropertyFunc func);
Property * FindProperty(Proxy * proxy, const gchar * name);
void SetPropertyFunc(Proxy * proxy, const gchar * name, PropertyFunc func);
void SetPropertyLazy(Proxy * proxy, const gchar * name, gboolean lazy);
void AddInitInstance(Proxy * proxy, InstanceInitFunc func);
void SetCreateInstance(Proxy * proxy, InstanceCreateFunc func);

void CheckNodeScalar(Node * node);
void CheckNodeVector(Node * node);
void CheckNodeObject(Node * node);

gint ModelGetInt(Node * node);
guint ModelGetUnsigned(Node * node);
gboolean ModelGetBool(Node * node);
gfloat ModelGetFloat(Node * node);
gdouble ModelGetDouble(Node * node);
gchar * ModelGetString(Node * node);
gunichar ModelGetChar(Node * node);
gint ModelGetEnum(Node * node);
CrowPackOptions ModelGetPackOptions(Node * node);
CrowPoint ModelGetPoint(Node * node);
GtkBorder ModelGetGtkBorder(Node * node);
GdkColor ModelGetColor(Node * node);
CrowEmitter * ModelGetEmitter(Node * node);
GObject * ModelGetObject(GuiLoader * loader, Node * node, GObject * old);
PtrArray * ModelGetObjectVector(GuiLoader * loader, Node * node);
void FreeObjectVector(PtrArray * array);
PtrArray * ModelGetStringVector(Node * node);
PtrArray * ModelGetEmitterVector(Node * node);

void PropertyInt(PropertyFuncData * data);
void PropertyUnsigned(PropertyFuncData * data);
void PropertyBool(PropertyFuncData * data);
void PropertyFloat(PropertyFuncData * data);
void PropertyDouble(PropertyFuncData * data);
void PropertyString(PropertyFuncData * data);
void PropertyChar(PropertyFuncData * data);
void PropertyEnum(PropertyFuncData * data);
void PropertyObject(PropertyFuncData * data);
void PropertyColor(PropertyFuncData * data);

void DrawGrid(GdkDrawable * window, GtkStyle * style, gboolean squares);

#endif
